@@ -23,9 +23,10 @@ PPI::Document - Object representation of a Perl document
   $Document->prune('PPI::Token::Comment');
   
   # Find all the named subroutines
-  my @subs = $Document->find( 
+  my $sub_nodes = $Document->find( 
   	sub { $_[1]->isa('PPI::Statement::Sub') and $_[1]->name }
   	);
+  my @sub_names = map { $_->name } @$sub_nodes;
   
   # Save the file
   $Document->save('My/Module.pm.stripped');
@@ -64,23 +65,28 @@ Document-specific.
 
 use strict;
 use base 'PPI::Node';
-use Carp                    ();
-use List::MoreUtils         ();
-use Params::Util            '_INSTANCE',
-                            '_SCALAR';
-use Digest::MD5             ();
-use PPI                     ();
-use PPI::Util               ();
-use PPI::Document::Fragment ();
-use overload 'bool'         => sub () { 1 };
-use overload '""'           => 'content';
+use Carp                          ();
+use List::MoreUtils               ();
+use Params::Util                  '_INSTANCE',
+                                  '_SCALAR';
+use Digest::MD5                   ();
+use PPI                           ();
+use PPI::Util                     ();
+use PPI::Exception::ParserTimeout ();
+use overload 'bool'               => sub () { 1 };
+use overload '""'                 => 'content';
+use constant HAS_ALARM            => (
+	$^O eq 'MSWin32' or $^O eq 'cygwin'
+	) ? 0 : 1;
 
 use vars qw{$VERSION $errstr};
 BEGIN {
-	$VERSION = '1.118';
+	$VERSION = '1.203';
 	$errstr  = '';
 }
 
+use PPI::Document::Fragment ();
+
 # Document cache
 my $CACHE = undef;
 
@@ -145,8 +151,15 @@ sub new {
 		return $self;
 	}
 
+	# Check constructor attributes
+	my $source  = shift;
+	my %attr    = @_;
+	my $timeout = delete $attr{timeout};
+	if ( $timeout and ! HAS_ALARM ) {
+		Carp::croak("This platform does not support PPI parser timeouts");
+	}
+
 	# Check the data source
-	my $source = shift;
 	if ( ! defined $source ) {
 		$class->_error("An undefined value was passed to PPI::Document::new");
 
@@ -167,30 +180,70 @@ sub new {
 
 			# Retrieve the document from the cache
 			my $document = $CACHE->get_document($source);
-			return $class->_setattr( $document, @_ ) if $document;
+			return $class->_setattr( $document, %attr ) if $document;
 
-			$document = PPI::Lexer->lex_source( $$source );
+			if ( $timeout ) {
+				eval {
+					local $SIG{ALRM} = sub { die "alarm\n" };
+					alarm( $timeout );
+					$document = PPI::Lexer->lex_source( $$source );
+					alarm( 0 );
+				};
+			} else {
+				$document = PPI::Lexer->lex_source( $$source );
+			}
 			if ( $document ) {
 				# Save in the cache
 				$CACHE->store_document( $document );
-				return $class->_setattr( $document, @_ );
+				return $class->_setattr( $document, %attr );
 			}
 		} else {
-			my $document = PPI::Lexer->lex_file( $source );
-			return $class->_setattr( $document, @_ ) if $document;
+			if ( $timeout ) {
+				eval {
+					local $SIG{ALRM} = sub { die "alarm\n" };
+					alarm( $timeout );
+					my $document = PPI::Lexer->lex_file( $source );
+					return $class->_setattr( $document, %attr ) if $document;
+					alarm( 0 );
+				};
+			} else {
+				my $document = PPI::Lexer->lex_file( $source );
+				return $class->_setattr( $document, %attr ) if $document;
+			}
 		}
 
 	} elsif ( _SCALAR($source) ) {
-		my $document = PPI::Lexer->lex_source( $$source );
-		return $class->_setattr( $document, @_ ) if $document;
+		if ( $timeout ) {
+			eval {
+				local $SIG{ALRM} = sub { die "alarm\n" };
+				alarm( $timeout );
+				my $document = PPI::Lexer->lex_source( $$source );
+				return $class->_setattr( $document, %attr ) if $document;
+				alarm( 0 );
+			}
+		} else {
+			my $document = PPI::Lexer->lex_source( $$source );
+			return $class->_setattr( $document, %attr ) if $document;
+		}
 
 	} else {
 		$class->_error("An unknown object or reference was passed to PPI::Document::new");
 	}
 
 	# Pull and store the error from the lexer
-	my $errstr = PPI::Lexer->errstr
-		|| "Unknown error returned by PPI::Lexer";
+	my $errstr;
+	if ( _INSTANCE($@, 'PPI::Exception::Timeout') ) {
+		$errstr = 'Timed out while parsing document';
+	} elsif ( _INSTANCE($@, 'PPI::Exception') ) {
+		$errstr = $@->message;
+	} elsif ( $@ ) {
+		$errstr = $@;
+		$errstr =~ s/\sat line\s.+$//;
+	} elsif ( PPI::Lexer->errstr ) {
+		$errstr = PPI::Lexer->errstr;
+	} else {
+		$errstr = "Unknown error parsing Perl document";
+	}
 	PPI::Lexer->_clear;
 	$class->_error( $errstr );
 }
@@ -318,10 +371,11 @@ or write to the file.
 
 sub save {
 	my $self = shift;
-	open( PPIOUTPUT, ">", $_[0] )    or return undef;
-	print PPIOUTPUT $self->serialize or return undef;
-	close PPIOUTPUT                  or return undef;
-	1;
+	local *FILE;
+	open( FILE, '>', $_[0] )    or return undef;
+	print FILE $self->serialize or return undef;
+	close FILE                  or return undef;
+	return 1;
 }
 
 =pod
@@ -672,6 +726,18 @@ sub scope { 1 }
 #####################################################################
 # PPI::Element Methods
 
+# Is the document complete.
+# Cascase to the last significant child
+sub complete {
+	my $self  = shift;
+	my $child = $self->schild(-1);
+	return !! (
+		$child
+		and
+		$child->complete
+	);
+}
+
 sub insert_before {
 	return undef;
 	# die "Cannot insert_before a PPI::Document";
@@ -768,7 +834,7 @@ L<PPI>, L<http://ali.as/>
 
 =head1 COPYRIGHT
 
-Copyright 2001 - 2006 Adam Kennedy.
+Copyright 2001 - 2008 Adam Kennedy.
 
 This program is free software; you can redistribute
 it and/or modify it under the same terms as Perl itself.
