@@ -61,7 +61,7 @@ use PPI::Document ();
 
 use vars qw{$VERSION $errstr};
 BEGIN {
-	$VERSION = '1.118';
+	$VERSION = '1.203';
 	$errstr  = '';
 }
 
@@ -232,11 +232,10 @@ sub _lex_document {
 			my $Statement = $_class->new( $Token ) or return undef;
 
 			# Move the lexing down into the statement
-			$self->_add_delayed( $Document )    or return undef;
-			$self->_lex_statement( $Statement ) or return undef;
-
-			# Add the completed Statement to the document
+			$self->_add_delayed( $Document )             or return undef;
 			$self->_add_element( $Document, $Statement ) or return undef;
+			$self->_lex_statement( $Statement )          or return undef;
+
 			next;
 		}
 
@@ -245,8 +244,8 @@ sub _lex_document {
 			# This should actually have a Statement instead
 			$self->_rollback( $Token );
 			my $Statement = PPI::Statement->new          or return undef;
-			$self->_lex_statement( $Statement )          or return undef;
 			$self->_add_element( $Document, $Statement ) or return undef;
+			$self->_lex_statement( $Statement )          or return undef;
 			next;
 		}
 
@@ -274,6 +273,23 @@ sub _lex_document {
 	# No error, it's just the end of file.
 	# Add any insignificant trailing tokens.
 	$self->_add_delayed( $Document );
+
+	# If the Tokenizer has any v6 blocks to attach, do so now.
+	# Checking once at the end is faster than adding a special
+	# case check for every statement parsed.
+	my $v6 = $self->{Tokenizer}->{v6};
+	if ( @$v6 ) {
+		my $includes = $Document->find( 'PPI::Statement::Include::Perl6' );
+		foreach my $include ( @$includes ) {
+			unless ( @$v6 ) {
+				my $errstr = "Failed to find a v6 section";
+				return $self->_error( $errstr );
+			}
+			$include->{perl6} = shift @$v6;
+		}
+	}
+
+	return 1;
 }
 
 
@@ -295,7 +311,7 @@ BEGIN {
 
 		# Loading and context statement
 		'package'  => 'PPI::Statement::Package',
-		'use'      => 'PPI::Statement::Include',
+		# 'use'      => 'PPI::Statement::Include',
 		'no'       => 'PPI::Statement::Include',
 		'require'  => 'PPI::Statement::Include',
 
@@ -367,12 +383,36 @@ sub _resolve_new_statement {
 
 	# Handle potential barewords for subscripts
 	if ( $Parent->isa('PPI::Structure::Subscript') ) {
-		if ( $class and $class->isa('PPI::Statement::Expression') ) {
-			# Still allowable in this context
-			return $class;
-		} else {
+		# Fast obvious case, just an expression
+		unless ( $class and $class->isa('PPI::Statement::Expression') ) {
 			return 'PPI::Statement::Expression';
 		}
+
+		# This is something like "my" or "our" etc... more subtle.
+		# Check if the next token is a closing curly brace.
+		# This means we are something like $h{my}
+		my $Next;
+		while ( $Next = $self->_get_token ) {
+			unless ( $Next->significant ) {
+				$self->_delay_element( $Next ) or return undef;
+				next;
+			}
+
+			# Found the next significant token.
+			# Is it a closing curly brace?
+			if ( $Next->content eq '}' ) {
+				$self->_rollback( $Next );
+				return 'PPI::Statement::Expression';
+			} else {
+				$self->_rollback( $Next );
+				return $class;
+			}
+		}
+
+		# End of file... this means it is something like $h{our
+		# which is probably going to be $h{our} ... I think
+		$self->_rollback( $Next );
+		return 'PPI::Statement::Expression';
 	}
 
 	# If it's a token in our list, use that class
@@ -420,6 +460,32 @@ sub _resolve_new_statement {
 		return 'PPI::Statement::Sub';
 	}
 
+	if ( $Token->content eq 'use' ) {
+		# Add a special case for "use v6" lines.
+		my $Next;
+		while ( $Next = $self->_get_token ) {
+			unless ( $Next->significant ) {
+				$self->_delay_element( $Next ) or return undef;
+				next;
+			}
+
+			# Found the next significant token.
+			# Is it a v6 use?
+			if ( $Next->content eq 'v6' ) {
+				$self->_rollback( $Next );
+				return 'PPI::Statement::Include::Perl6';
+			} else {
+				$self->_rollback( $Next );
+				return 'PPI::Statement::Include';
+			}
+		}
+
+		# End of file... this means it is an incomplete use
+		# line, just treat it as a normal include.
+		$self->_rollback( $Next );
+		return 'PPI::Statement::Include';
+	}
+
 	# If our parent is a Condition, we are an Expression
 	if ( $Parent->isa('PPI::Structure::Condition') ) {
 		return 'PPI::Statement::Expression';
@@ -494,10 +560,8 @@ sub _lex_statement {
 
 		# Move the lexing down into the Structure
 		$self->_add_delayed( $Statement )   or return undef;
-		$self->_lex_structure( $Structure ) or return undef;
-
-		# Add the completed Structure to the statement
 		$self->_add_element( $Statement, $Structure ) or return undef;
+		$self->_lex_structure( $Structure ) or return undef;
 	}
 
 	# Was it an error in the tokenizer?
@@ -774,10 +838,10 @@ BEGIN {
 # Given a parent element, and a token which will open a structure, determine
 # the class that the structure should be.
 sub _resolve_new_structure {
-	my ($self, $Parent, $Token) = @_;
 	# my $self   = shift;
 	# my $Parent = _INSTANCE(shift, 'PPI::Node')             or die "Bad param 1";
 	# my $Token  = _INSTANCE(shift, 'PPI::Token::Structure') or die "Bad param 2";
+	my ($self, $Parent, $Token) = @_;
 
 	return $self->_resolve_new_structure_round ($Parent) if $Token->content eq '(';
 	return $self->_resolve_new_structure_square($Parent) if $Token->content eq '[';
@@ -788,9 +852,9 @@ sub _resolve_new_structure {
 # Given a parent element, and a ( token to open a structure, determine
 # the class that the structure should be.
 sub _resolve_new_structure_round {
-	my ($self, $Parent) = @_;
 	# my $self   = shift;
 	# my $Parent = _INSTANCE(shift, 'PPI::Node') or die "Bad param 1";
+	my ($self, $Parent) = @_;
 
 	# Get the last significant element in the parent
 	my $Element = $Parent->schild(-1);
@@ -819,9 +883,9 @@ sub _resolve_new_structure_round {
 # Given a parent element, and a [ token to open a structure, determine
 # the class that the structure should be.
 sub _resolve_new_structure_square {
-	my ($self, $Parent) = @_;
 	# my $self   = shift;
 	# my $Parent = _INSTANCE(shift, 'PPI::Node') or die "Bad param 1";
+	my ($self, $Parent) = @_;
 
 	# Get the last significant element in the parent
 	my $Element = $Parent->schild(-1);
@@ -849,6 +913,26 @@ sub _resolve_new_structure_square {
 	'PPI::Structure::Constructor';
 }
 
+use vars qw{%CURLY_CLASSES};
+BEGIN {
+	# Keyword -> Structure class maps
+	%CURLY_CLASSES = (
+		# Blocks
+		'sub'  => 'PPI::Structure::Block',
+		'grep' => 'PPI::Structure::Block',
+		'map'  => 'PPI::Structure::Block',
+		'sort' => 'PPI::Structure::Block',
+		'do'   => 'PPI::Structure::Block',
+
+		# Hash constructors
+		'='    => 'PPI::Structure::Constructor',
+		'||='  => 'PPI::Structure::Constructor',
+		','    => 'PPI::Structure::Constructor',
+		'=>'   => 'PPI::Structure::Constructor',
+		
+		);
+}
+
 # Given a parent element, and a { token to open a structure, determine
 # the class that the structure should be.
 sub _resolve_new_structure_curly {
@@ -858,10 +942,11 @@ sub _resolve_new_structure_curly {
 
 	# Get the last significant element in the parent
 	my $Element = $Parent->schild(-1);
+	my $content = $Element ? $Element->content : '';
 
 	# Is this a subscript, like $foo[1] or $foo{expr}
 	if ( $Element ) {
-		if ( $Element->isa('PPI::Token::Operator') and $Element->content eq '->' ) {
+		if ( $content eq '->' and $Element->isa('PPI::Token::Operator') ) {
 			# $foo->{}
 			$Element->{_dereference} = 1;
 			return 'PPI::Structure::Subscript';
@@ -870,10 +955,14 @@ sub _resolve_new_structure_curly {
 			# $foo[]{}
 			return 'PPI::Structure::Subscript';
 		}
-		if ( $Element->isa('PPI::Token::Symbol') and $Element->content =~ /^(?:\$|\@)/ ) {
+		if ( $Element->isa('PPI::Token::Symbol') and $content =~ /^(?:\$|\@)/ ) {
 			# $foo{}, @foo{}
 			return 'PPI::Structure::Subscript';
 		}
+		if ( $CURLY_CLASSES{$content} ) {
+			# Known type
+			return $CURLY_CLASSES{$content};
+		}
 	}
 
 	# Are we in a compound statement
@@ -882,11 +971,33 @@ sub _resolve_new_structure_curly {
 		return 'PPI::Structure::Block';
 	}
 
-	# Unless we are at the start of the line, everything should be a block
+	# Are we the second argument of use
+	if ( $Parent->isa('PPI::Statement::Include') ) {
+		if ( $Parent->schildren == 2 ) {
+			# This is something like use constant { ... };
+			return 'PPI::Structure::Constructor';
+		}
+	}
+
+	# Unless we are at the start of the statement, everything else should be a block
+	### FIXME This is possibly a bad choice, but will have to do for now.
 	if ( $Element ) {
 		return 'PPI::Structure::Block';
 	}
 
+	# Special case: Are we the param of a core function
+	# i.e. map({ $_ => 1 } @foo)
+	if (
+		$Parent->isa('PPI::Statement')
+		and
+		_INSTANCE($Parent->parent, 'PPI::Structure::List')
+	) {
+		my $function = $Parent->parent->parent->schild(-2);
+		if ( $function and $function->content =~ /^(?:map|grep|sort)$/ ) {
+			return 'PPI::Structure::Block';
+		}
+	}
+
 	# We need to scan ahead.
 	my $Next;
 	my $position = 0;
@@ -900,14 +1011,14 @@ sub _resolve_new_structure_curly {
 		# If a closing curly, this is an anonymous hash-ref
 		if ( ++$position == 1 and $Next->content eq '}' ) {
 			$self->_buffer( splice( @delayed ), $Next );
-			return 'PPI::Structure::Subscript';
+			return 'PPI::Structure::Constructor';
 		}
 
 		# If it contains a => as the second thing,
 		# this is also an anonymous hash-ref.
 		if ( $position == 2 and $Next->content eq '=>' ) {
 			$self->_buffer( splice( @delayed ), $Next );
-			return 'PPI::Structure::Subscript';
+			return 'PPI::Structure::Constructor';
 		}
 
 		# We only check the first two, then default to block
@@ -950,40 +1061,22 @@ sub _lex_structure {
 
 			# Determine the class for the Statement and create it
 			my $_class = $self->_resolve_new_statement($Structure, $Token) or return undef;
-			my $Statement = $_class->new( $Token ) or return undef;
+			my $Statement = $_class->new( $Token )        or return undef;
 
 			# Move the lexing down into the Statement
-			$self->_lex_statement( $Statement ) or return undef;
-
-			# Add the completed statement to our elements
 			$self->_add_element( $Structure, $Statement ) or return undef;
+			$self->_lex_statement( $Statement )           or return undef;
+
 			next;
 		}
 
 		# Is this the opening of another structure directly inside us?
 		if ( $Token->__LEXER__opens ) {
-			### FIXME - Now, we really shouldn't be creating Structures
-			###         inside of Structures. There really should be an
-			###         Statement::Expression in here somewhere.
-			# Determine the class for the structure and create it
-			my $_class = $self->_resolve_new_structure($Structure, $Token) or return undef;
-			if ( $_class eq 'PPI::Structure::List' ) {
-				# This should actually have a Statement
-				$self->_rollback( $Token );
-				my $Statement = PPI::Statement->new           or return undef;
-				$self->_add_delayed( $Structure )             or return undef;
-				$self->_lex_statement( $Statement )           or return undef;
-				$self->_add_element( $Structure, $Statement ) or return undef;
-				next;
-			}
-			my $Structure2 = $_class->new( $Token ) or return undef;
-
-			# Move the lexing down into the Structure
-			$self->_add_delayed( $Structure ) or return undef;
-			$self->_lex_structure( $Structure2 ) or return undef;
-
-			# Add the completed Structure to the statement
-			$self->_add_element( $Structure, $Structure2 ) or return undef;
+			# Rollback the Token, and recurse into the statement
+			$self->_rollback( $Token );
+			my $Statement = PPI::Statement->new          or return undef;
+			$self->_add_element( $Structure, $Statement ) or return undef;
+			$self->_lex_statement( $Statement )          or return undef;
 			next;
 		}
 
@@ -1096,7 +1189,7 @@ sub _add_delayed {
 
 	# Clear the delayed elements
 	$self->{delayed} = [];
-	1;
+	return 1;
 }
 
 # Rollback the delayed tokens, plus any passed. Once all the tokens
@@ -1199,7 +1292,7 @@ Adam Kennedy E<lt>adamk@cpan.orgE<gt>
 
 =head1 COPYRIGHT
 
-Copyright 2001 - 2006 Adam Kennedy.
+Copyright 2001 - 2008 Adam Kennedy.
 
 This program is free software; you can redistribute
 it and/or modify it under the same terms as Perl itself.
